package k8s

import (
	"context"
	"fmt"

	"git.earthnet.ch/simon.beck/kopia-k8s/logger"
	snapv1 "github.com/kubernetes-csi/external-snapshotter/client/v7/apis/volumesnapshot/v1"
	"github.com/urfave/cli/v2"
	corev1 "k8s.io/api/core/v1"
	"k8s.io/apimachinery/pkg/api/errors"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/utils/ptr"
	"sigs.k8s.io/controller-runtime/pkg/client"
)

// +kubebuilder:rbac:groups="",resources=persistentvolumeclaims,verbs=get;list;watch;create;update;patch;delete
// +kubebuilder:rbac:groups=snapshot.storage.k8s.io,resources=volumesnapshotclasses,verbs=get;list;watch
// +kubebuilder:rbac:groups=snapshot.storage.k8s.io,resources=volumesnapshots,verbs=get;list;watch;create;update;patch;delete

// MountedPVC describes a PVC and the POD it belongs to.
type MountedPVC struct {
	Pod *corev1.Pod
	PVC *corev1.PersistentVolumeClaim
}

// BackupPVCList contains two lists of PVCs.
// One list contains the PVCs that are currently mounted.
// The other contains the PVCs that aren't mounted.
type BackupPVCList struct {
	// MountedPVCs consists of the key <pvcname:namespace>.
	// This helps with deduplicating the PVCs afterwards.
	MountedPVCs   map[string]MountedPVC
	UnmountedPVCs *corev1.PersistentVolumeClaimList
}

// ListEligiblePVCs will list all PVCs that fullfill certain constraints.
func ListEligiblePVCs(cliCtx *cli.Context, k8sClient client.Client) (*BackupPVCList, error) {
	log := logger.AppLogger(cliCtx.Context).WithName("PVCLister")

	pods, err := listPodsWithPVCs(cliCtx, k8sClient)
	if err != nil {
		return nil, err
	}

	backupList := &BackupPVCList{MountedPVCs: map[string]MountedPVC{}}

	for _, tmp := range pods.Items {
		pod := tmp
		for _, volume := range pod.Spec.Volumes {
			if volume.PersistentVolumeClaim != nil {
				pvc, err := getPVCFromClaimSource(cliCtx, k8sClient, volume.PersistentVolumeClaim, &pod)
				backupListKey := fmt.Sprintf("%s:%s", pvc.Name, pvc.Namespace)
				if err != nil {
					return nil, fmt.Errorf("could not get pvc for pod %s: %w", pod.Name, err)
				}
				backupList.MountedPVCs[backupListKey] = MountedPVC{Pod: &pod, PVC: pvc}
				log.V(1).Info("found pod and pvc", "podname", pod.Name, "pvcname", pvc.Name, "namespace", pod.Namespace)
			}
		}
	}

	allPVCs := corev1.PersistentVolumeClaimList{}
	err = k8sClient.List(cliCtx.Context, &allPVCs)

	backupList.UnmountedPVCs = &corev1.PersistentVolumeClaimList{}
	for _, pvc := range allPVCs.Items {
		pvcKey := fmt.Sprintf("%s:%s", pvc.Name, pvc.Namespace)
		if _, ok := backupList.MountedPVCs[pvcKey]; !ok {
			backupList.UnmountedPVCs.Items = append(backupList.UnmountedPVCs.Items, pvc)
			log.V(1).Info("found unmounted PVC", "pvcname", pvc.Name, "namespace", pvc.Namespace)
		}
	}

	return backupList, err
}

func getPVCFromClaimSource(cliCtx *cli.Context, k8sClient client.Client, PVCSource *corev1.PersistentVolumeClaimVolumeSource, pod *corev1.Pod) (*corev1.PersistentVolumeClaim, error) {
	objectKey := client.ObjectKey{
		Namespace: pod.Namespace,
		Name:      PVCSource.ClaimName,
	}
	pvc := &corev1.PersistentVolumeClaim{}
	err := k8sClient.Get(cliCtx.Context, objectKey, pvc)
	return pvc, err
}

func snapshotPVC(ctx context.Context, k8sClient client.Client, namespace, snapClass, claim, podName string) (*snapv1.VolumeSnapshot, error) {
	snap := &snapv1.VolumeSnapshot{
		ObjectMeta: metav1.ObjectMeta{
			Name:      podName,
			Namespace: namespace,
		},
		Spec: snapv1.VolumeSnapshotSpec{
			VolumeSnapshotClassName: ptr.To(snapClass),
			Source: snapv1.VolumeSnapshotSource{
				PersistentVolumeClaimName: ptr.To(claim),
			},
		},
	}

	err := k8sClient.Create(ctx, snap)
	if err != nil && errors.IsAlreadyExists(err) {
		return snap, nil
	}

	return snap, err
}

func DeleteSnapshot(ctx context.Context, k8sClient client.Client, namespace, podName string) error {
	snap := &snapv1.VolumeSnapshot{
		ObjectMeta: metav1.ObjectMeta{
			Name:      podName,
			Namespace: namespace,
		},
	}

	return k8sClient.Delete(ctx, snap)
}

func DeletePVC(ctx context.Context, k8sClient client.Client, namespace, podName string) error {
	pvc := &corev1.PersistentVolumeClaim{
		ObjectMeta: metav1.ObjectMeta{
			Name:      podName,
			Namespace: namespace,
		},
	}

	return k8sClient.Delete(ctx, pvc)
}

func PVCFromSnapshot(ctx context.Context, k8sClient client.Client, namespace, snapName, podName string, src *corev1.PersistentVolumeClaim) (*corev1.PersistentVolumeClaim, error) {
	pvc := &corev1.PersistentVolumeClaim{
		ObjectMeta: metav1.ObjectMeta{
			Name:      podName,
			Namespace: namespace,
		},
		Spec: corev1.PersistentVolumeClaimSpec{
			AccessModes: []corev1.PersistentVolumeAccessMode{
				corev1.ReadWriteOnce,
			},
			DataSource: &corev1.TypedLocalObjectReference{
				Name:     snapName,
				Kind:     "VolumeSnapshot",
				APIGroup: ptr.To("snapshot.storage.k8s.io"),
			},
		},
	}

	pvc.Spec.Resources.Requests = src.Status.Capacity

	err := k8sClient.Create(ctx, pvc)
	if err != nil && errors.IsAlreadyExists(err) {
		return nil, nil
	}

	return pvc, err
}

func getSnapshotClass(ctx context.Context, pvc *corev1.PersistentVolumeClaim, k8sClient client.Client) (string, error) {
	provider := pvc.GetAnnotations()["volume.beta.kubernetes.io/storage-provisioner"]
	if provider == "" {
		return "", fmt.Errorf("cannot find a snapshot class, falling back to normal backup")
	}

	vsc := &snapv1.VolumeSnapshotClassList{}
	err := k8sClient.List(ctx, vsc)
	if err != nil {
		return "", fmt.Errorf("cannot get snapshot classes, falling back to normal backup: %w", err)
	}

	for _, class := range vsc.Items {
		if class.Driver == provider {
			return class.GetName(), nil
		}
	}

	return "", fmt.Errorf("cannot find a suitable snapshotclass, falling back to normal backup")
}
